<?php

namespace App\Http\Controllers\User;

use App\Constants\Status;
use App\Http\Controllers\Controller;
use App\Models\Bid;
use App\Models\Category;
use App\Models\Contract;
use App\Models\Item;
use App\Models\NftActivity;
use App\Models\User;
use App\Models\Wishlist;
use Illuminate\Http\Request;

class ItemController extends Controller
{
    public function allList(Request $request)
    {
        $pageTitle = 'Marketplace';
        $items     = Item::where('items.status', Status::ITEM_LISTED)->searchable(['nft:name']);

        if ($request->min) {
            $items->where('amount', '>=', $request->min);
        }

        if ($request->max) {
            $items->where('amount', '<=', $request->max);
        }

        if ($request->category) {
            $items->whereHas('nft', function ($nft) use ($request) {
                $nft->listed()->where('category_id', $request->category)->whereHas('category', function ($category) {
                    $category->active();
                });
            });
        }

        if ($request->status && $request->status != 'all') {
            $status = $request->status == 'auction' ? Status::YES : Status::NO;
            $items->where('is_auction', $status);
        }

        if ($request->sort) {
            $sortBy = explode('-', $request->sort);
            if ($sortBy[0] == 'name') {
                $items->join('nfts', 'items.nft_id', '=', 'nfts.id')
                    ->orderBy('nfts.name', $sortBy[1]);
            } elseif ($sortBy[0] == 'price') {
                $items->orderBy('amount', $sortBy[1]);
            } else {
                $items->orderBy('created_at', $sortBy[1]);
            }
        }

        $items = $items->with(['nft', 'user', 'bids.user'])->withCount(['wishlists' => function ($wishlist) {
            $wishlist->where('user_id', auth()->id());
        }])->paginate(getPaginate(12));

        if ($request->ajax()) {
            return view('Template::user.marketplace.content', compact('items'));
        }

        $categories          = Category::active()->get();
        $contracts           = Contract::latest()->get();
        $marketplace         = $contracts->where('act', 'marketplace')->first();
        $marketplaceContract = @$marketplace->contract;
        $nft                 = $contracts->where('act', 'nft')->first();
        $nftContract         = @$nft->contract;

        return view('Template::user.marketplace.all_list', compact('pageTitle', 'items', 'categories'));
    }

    public function listed()
    {
        return $this->itemData('listed');
    }

    public function unlisted()
    {
        return $this->itemData('unlisted');
    }
    public function sold()
    {
        return $this->itemData('sold');
    }

    private function itemData($scope = null)
    {
        if ($scope) {
            $items     = Item::$scope();
            $pageTitle = ucFirst($scope) . ' Items';
        } else {
            $items     = Item::query();
            $pageTitle = 'All Items';
        }

        $items = $items->where('user_id', auth()->id())->with('nft', 'bids.user')->withCount(['wishlists' => function ($wishlist) {
            $wishlist->where('user_id', auth()->id());
        }])->whereHas('nft.category', function ($category) {
            $category->active();
        })->orderBy('id', 'desc')->paginate(getPaginate());

        $contracts           = Contract::latest()->get();
        $marketplace         = $contracts->where('act', 'marketplace')->first();
        $marketplaceContract = @$marketplace->contract;
        $nft                 = $contracts->where('act', 'nft')->first();
        $nftContract         = $nft->contract;
        $user                = auth()->user();

        return view('Template::user.marketplace.list', compact('pageTitle', 'nftContract', 'marketplaceContract', 'items', 'user'));
    }

    public function unclaimBid()
    {
        $pageTitle = 'Unclaim Bid';
        $wishlists = Wishlist::with('item.bids.user', 'item.user', 'item.nft')->where('user_id', auth()->id())->paginate(getPaginate());

        $items = Item::where('status', Status::ITEM_SOLD)->where('is_auction', Status::YES)
            ->with(['bids' => function ($query) {
                $query->orderBy('amount', 'desc');
            }])
            ->whereHas('bids', function ($bid) {
                $bid->where('user_id', auth()->id())->where('is_winner', 0);
            })->paginate(getPaginate());

        return view('Template::user.unclaim_bid_item', compact('pageTitle', 'items'));
    }

    public function purchased(Request $request, $itemId)
    {
        $request->validate([
            'transactionHash' => 'required|string',
        ]);
        $item = Item::listed()->with(['bids', 'nft', 'user'])->where('itemId', $itemId)->first();

        if (!$item || $item->status != Status::ITEM_LISTED) {
            return false;
        }

        $item->status = Status::ITEM_SOLD;
        $item->save();

        Wishlist::where('item_id', $item->id)->delete();

        $nft          = $item->nft;
        $nft->user_id = auth()->id();
        $nft->status  = Status::NFT_ACTIVE;
        $nft->save();

        $nftActivity                   = new NftActivity();
        $nftActivity->nft_id           = $item->nft_id;
        $nftActivity->amount           = $item->amount;
        $nftActivity->transaction_hash = $request->transactionHash;
        $nftActivity->from_wallet      = $item->user->wallet;
        $nftActivity->to_wallet        = auth()->user()->wallet;
        $nftActivity->type             = 'purchased';
        $nftActivity->description      = 'NFT Purchased';
        $nftActivity->save();

        notify(auth()->user(), 'PURCHASED_ITEM', [
            'nft_name'             => $item->nft->name,
            'amount'               => showAmount($item->amount),
            'currency'             => marketplaceCurrency(),
            'transaction_hash'     => showAddress($request->transactionHash),
            'transaction_hash_url' => blockExplorer() . '/tx/' . $request->transactionHash,
        ]);

        notify($item->user, 'ITEM_SOLD', [
            'nft_name'             => $item->nft->name,
            'amount'               => showAmount($item->amount),
            'currency'             => marketplaceCurrency(),
            'transaction_hash'     => showAddress($request->transactionHash),
            'transaction_hash_url' => blockExplorer() . '/tx/' . $request->transactionHash,
        ]);

        return true;
    }

    public function bidNow(Request $request, $itemId)
    {
        $request->validate([
            'amount' => 'required|string|gt:0',
        ]);
        $item = Item::listed()->with('bids.user', 'nft')->where('itemId', $itemId)->where('user_id', '!=', auth()->id())->firstOrFail();
        if ($item->status != Status::ITEM_LISTED) {
            return false;
        }

        $user   = auth()->user();
        $exists = $item->bids->where('user_id', $user->id)->first();
        if ($exists) {
            return false;
        }

        $lowestBids = $item->bids;

        $bid          = new Bid();
        $bid->item_id = $item->id;
        $bid->user_id = $user->id;
        $bid->amount  = $request->amount;
        $bid->save();

        notify($item->user, 'ITEM_BID', [
            'nft_name' => $item->nft->name,
            'amount'   => showAmount($request->amount),
            'currency' => marketplaceCurrency(),
            'wallet'   => $user->username ?? showAddress($user->wallet),
        ]);

        foreach ($lowestBids as $bid) {
            notify($bid->user, 'LOWER_BID', [
                'nft_name'    => $item->nft->name,
                'your_amount' => showAmount($bid->amount),
                'amount'      => showAmount($request->amount),
                'currency'    => marketplaceCurrency(),
                'wallet'      => $user->username ?? showAddress($user->wallet),
            ]);
        }

        return true;
    }
    public function bidUpdate(Request $request, $itemId)
    {
        $request->validate([
            'amount' => 'required|string|gt:0',
        ]);
        $item = Item::listed()->with('bids.user')->where('itemId', $itemId)->where('user_id', '!=', auth()->id())->firstOrFail();
        if ($item->status != Status::ITEM_LISTED) {
            return false;
        }

        $user = auth()->user();
        $bid  = $item->bids->where('user_id', $user->id)->first();
        if (!$bid) {
            return false;
        }

        $bid->amount = $bid->amount + $request->amount;
        $bid->save();

        notify($item->user, 'ITEM_BID_UPDATE', [
            'nft_name'     => $item->nft->name,
            'amount'       => showAmount($request->amount),
            'total_amount' => showAmount($bid->amount),
            'currency'     => marketplaceCurrency(),
            'wallet'       => $user->username ?? showAddress($user->wallet),
        ]);

        $lowestBids = $item->bids->where('user_id', '!=', $user->id);

        foreach ($lowestBids as $lowerBid) {
            notify($lowerBid->user, 'LOWER_BID', [
                'nft_name'    => $item->nft->name,
                'your_amount' => showAmount($lowerBid->amount),
                'amount'      => showAmount($bid->amount),
                'currency'    => marketplaceCurrency(),
                'wallet'      => $user->username ?? showAddress($user->wallet),
            ]);
        }

        return true;
    }

    public function bidClaim($itemId)
    {
        $item          = Item::listed()->with('bids')->where('itemId', $itemId)->where('user_id', '!=', auth()->id())->firstOrFail();
        $highestBid    = $item->bids->sortByDesc('amount')->first();
        $myPreviousBid = $item->bids->where('user_id', auth()->id())->first();
        if ($highestBid->id == $myPreviousBid->id) {
            return false;
        }

        return $myPreviousBid->delete();
    }

    public function unlistItem($itemId)
    {
        $item = Item::where('user_id', auth()->id())->where('itemId', $itemId)->first();

        if (!$item || $item->status != Status::ITEM_LISTED) {
            return false;
        }

        $item->status = Status::ITEM_UNLISTED;
        $item->save();

        $nft         = $item->nft;
        $nft->status = Status::NFT_ACTIVE;
        $nft->save();

        $nftActivity                   = new NftActivity();
        $nftActivity->nft_id           = $item->nft_id;
        $nftActivity->transaction_hash = '#';
        $nftActivity->from_wallet      = '';
        $nftActivity->to_wallet        = auth()->user()->wallet;
        $nftActivity->type             = 'unlist';
        $nftActivity->description      = 'NFT unlisted from marketplace';
        $nftActivity->save();
        return true;
    }

    public function nftActivity(Request $request)
    {
        $nftActivities = NftActivity::where('nft_id', $request->nft_id);

        if ($request->activity_type != 'all') {
            $nftActivities->where('type', $request->activity_type);
        }

        if ($request->last_id) {
            $nftActivities->where('id', '<', $request->last_id);
        }

        $nftActivities = $nftActivities->whereHas('nft', function ($nft) {
            $nft->where('status', Status::NFT_LISTED)->orWhere('status', Status::NFT_ACTIVE);
        })->orderBy('id', 'desc')->limit(10)->get();

        return view('Template::partials.nft_activity_list', compact('nftActivities'));
    }
}
