<?php

namespace App\Http\Controllers\User;

use App\Constants\Status;
use App\Http\Controllers\Controller;
use App\Lib\CurlRequest;
use App\Lib\NFTStorage;
use App\Models\AdminNotification;
use App\Models\Bid;
use App\Models\Category;
use App\Models\Contract;
use App\Models\Item;
use App\Models\Nft;
use App\Models\NftActivity;
use App\Models\NftSize;
use App\Models\NftView;
use App\Models\Page;
use App\Rules\FileTypeValidate;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Validator;
use stdClass;

class NftController extends Controller
{
    public function mintForm()
    {
        $pageTitle  = 'Mint NFT';
        $categories = Category::active()->get();
        $nftSizes   = NftSize::get();
        return view('Template::user.nft.mint', compact('pageTitle', 'categories', 'nftSizes'));
    }

    public function all()
    {
        return $this->nftData();
    }

    public function active()
    {
        return $this->nftData('active');
    }

    public function listed()
    {
        return $this->nftData('listed');
    }

    public function burned()
    {
        return $this->nftData('burned');
    }

    private function nftData($scope = null)
    {
        $nfts      = Nft::query();
        $pageTitle = 'All Nft';
        if ($scope) {
            $nfts      = $nfts->$scope();
            $pageTitle = ucFirst($scope) . ' Nft';
        }

        $nfts = $nfts->where('user_id', auth()->id())->with(['items' => function ($item) {
            $item->latest('id');
        }])->whereHas('category', function ($category) {
            $category->active();
        })->orderBy('id', 'desc')->paginate(getPaginate());
        $user = auth()->user();

        return view('Template::user.nft.list', compact('pageTitle', 'nfts', 'user'));
    }

    public function nftList(Request $request)
    {
        $pageTitle = 'List Nft';
        $nfts      = Nft::latest()->where('user_id', auth()->id());

        if ($request->type == 'listed') {
            $nfts->where('status', Status::NFT_LISTED);
        } elseif ($request->type == 'active') {
            $nfts->where('status', Status::NFT_ACTIVE);
        } elseif ($request->type == 'burned') {
            $nfts->where('status', Status::NFT_BURNED);
        }

        $nfts = $nfts->whereHas('category', function ($category) {
            $category->active();
        })->with('items')->get();

        $contracts           = Contract::latest()->get();
        $marketplace         = $contracts->where('act', 'marketplace')->first();
        $marketplaceContract = $marketplace->contract;
        $nft                 = $contracts->where('act', 'nft')->first();
        $nftContract         = $nft->contract;
        $user                = auth()->user();

        return view('Template::user.nft.list', compact('pageTitle', 'nftContract', 'marketplaceContract', 'nfts', 'user'));
    }
    public function uploadToIpfs(Request $request)
    {
        $nftSizes       = NftSize::get();
        $validationRule = [
            'name'        => 'required|string|max:255',
            'category'    => 'required|integer',
            'file'        => ['required', new FileTypeValidate($nftSizes->pluck('file_extension')->reject(fn ($extension) => $extension === 'gif')->map(fn ($extension) => $extension)->toArray())],
            'animation'   => ['nullable', 'image', new FileTypeValidate(['gif'])],
            'thumb'       => ['nullable', 'image', new FileTypeValidate(['jpg', 'jpeg', 'png'])],
            'description' => 'required|string',
            'custom'      => 'nullable|array',
            'custom.*'    => 'nullable|array',
            'custom.*.*'  => 'nullable|string',
        ];

        if (gs('agree')) {
            $validationRule['agree'] = 'required';
        }

        $validator = Validator::make($request->all(), $validationRule);

        if ($validator->fails()) {
            return response()->json(['status' => 'error', 'message' => $validator->errors()]);
        }

        $fileSize      = $request->file->getSize() / 1024;
        $fileExtension = strtolower($request->file->getClientOriginalExtension());
        $fileMaxSize   = $nftSizes->where('file_extension', $fileExtension)->first()->max_size;

        if (($fileExtension == 'mp3' || $fileExtension == 'mp4') && !$request->thumb) {
            return response()->json(['status' => 'error', 'message' => 'Thumb is required']);
        }

        if ($fileSize > $fileMaxSize) {
            return response()->json(['status' => 'error', 'message' => strtoupper($fileExtension) . " file size not greater than $fileMaxSize KB"]);
        }

        if ($request->animation) {
            $fileSize    = $request->animation->getSize() / 1024;
            $fileMaxSize = $nftSizes->where('file_extension', 'gif')->first()->max_size;
            if ($fileSize > $fileMaxSize) {
                return response()->json(['status' => 'error', 'message' => "GIF file size not greater than $fileMaxSize KB"]);
            }
        }

        if ($request->thumb) {
            $fileSize      = $request->thumb->getSize() / 1024;
            $fileExtension = strtolower($request->thumb->getClientOriginalExtension());
            $fileMaxSize   = $nftSizes->where('file_extension', $fileExtension)->first()->max_size;

            if ($fileSize > $fileMaxSize) {
                return response()->json(['status' => 'error', 'message' => strtoupper($fileExtension) . " file size not greater than $fileMaxSize KB"]);
            }
        }

        $category = Category::where('status', Status::ENABLE)->find($request->category);

        if (!$category) {
            return response()->json(['status' => 'error', 'message' => 'Category not found']);
        }

        $nftFileUri      = null;
        $nftAnimationUri = null;

        if ($request->hasFile('file')) {
            $nftFileUri = NFTStorage::upload($request->file);
        }

        if (!$nftFileUri) {
            $adminNotification            = new AdminNotification();
            $adminNotification->user_id   = auth()->id();
            $adminNotification->title     = 'NFT Storage Pinata API is error. User cannot upload NFT file';
            $adminNotification->click_url = urlPath('admin.setting.nft.storage');
            $adminNotification->save();
            return response()->json(['status' => 'error', 'message' => "File Server Error"]);
        }

        if ($request->hasFile('animation')) {
            $nftAnimationUri = NFTStorage::upload($request->animation);
        }

        if ($request->hasFile('thumb')) {
            $nftAnimationUri = NFTStorage::upload($request->thumb);
        }

        $customData = new stdClass();
        foreach (@$request->custom['key'] ?? [] as $index => $key) {
            $key   = titleToKey($key);
            $value = @$request->custom['value'][$index];
            if ($key && $value) {
                $customData->$key = $value;
            }
        }

        $metadata = [
            'name'          => $request->name,
            'description'   => $request->description,
            'file_url'      => $nftFileUri,
            'external_url'  => route('home'),
            'animation_url' => $nftAnimationUri,
            'custom_fields' => $customData,
        ];
        $metadataUri = NFTStorage::upload($metadata, true);

        if (!$metadataUri) {
            $adminNotification            = new AdminNotification();
            $adminNotification->user_id   = auth()->id();
            $adminNotification->title     = 'NFT Storage Pinata API is error. User cannot upload NFT metadata';
            $adminNotification->click_url = urlPath('admin.setting.nft.storage');
            $adminNotification->save();
            return response()->json(['status' => 'error', 'message' => "File Server Error"]);
        }

        return response()->json([
            "uri" => $metadataUri,
        ]);
    }
    public function storeMint(Request $request)
    {
        $request->validate([
            'ifps_uri'         => 'required|string',
            'token_id'         => 'required|string',
            'category_id'      => 'required|string',
            'file_type'        => 'required|in:audio,video,image',
            'transaction_hash' => 'required|string',
        ]);

        $response = CurlRequest::curlContent($request->ifps_uri);
        $nftData  = json_decode($response);
        if (!@$nftData?->name) {
            $response = CurlRequest::curlContent($request->ifps_uri);
            $nftData  = json_decode($response);
            // try second time if it failed to fetch data
            // multiple times faced this issues
            // it is a pinata server issues for free users
        }
        if ($response) {
            $contract = Contract::nft()->first();

            $nft          = new Nft();
            $nft->user_id = auth()->id();

            $nft->category_id      = $request->category_id;
            $nft->ipfs_uri         = $request->ifps_uri;
            $nft->name             = @$nftData?->name;
            $nft->description      = @$nftData?->description;
            $nft->custom_data      = @$nftData?->custom_fields;
            $nft->file             = @$nftData?->file_url;
            $nft->file_type        = $request->file_type;
            $nft->chain            = marketplaceNetwork()->chain;
            $nft->animation        = @$nftData?->animation_url;
            $nft->transaction_hash = $request->transaction_hash;
            $nft->contract_address = $contract->contract;
            $nft->token_id         = $request->token_id;
            $nft->save();

            $nftActivity                   = new NftActivity();
            $nftActivity->nft_id           = $nft->id;
            $nftActivity->transaction_hash = @$request->transaction_hash;
            $nftActivity->from_wallet      = 0;
            $nftActivity->to_wallet        = auth()->user()->wallet;
            $nftActivity->type             = 'mint';
            $nftActivity->description      = 'Nft is minted';
            $nftActivity->save();
            return true;
        }
        return false;
    }
    public function storeList(Request $request)
    {
        $request->validate([
            'nftId'                 => 'required|exists:nfts,id',
            'amount'                => 'required|string',
            'auctionDateTimeInUnix' => 'required|string',
            'itemId'                => 'required|string',
        ]);

        $nft = Nft::where('id', $request->nftId)->with('items')->whereHas('category', function ($category) {
            $category->active();
        })->first();

        if ($nft->status != Status::NFT_ACTIVE) {
            return false;
        }

        $nft->status = Status::NFT_LISTED;
        $nft->save();

        $item                 = new Item();
        $item->nft_id         = $request->nftId;
        $item->user_id        = auth()->id();
        $item->amount         = $request->amount;
        $item->auction_end_at = $request->auctionDateTimeInUnix;
        $item->itemId         = $request->itemId;
        if ($item->auction_end_at > 0) {
            $item->is_auction = Status::YES;
        }

        $item->save();
        $nftActivity                   = new NftActivity();
        $nftActivity->nft_id           = $nft->id;
        $nftActivity->amount           = $request->amount;
        $nftActivity->transaction_hash = '#';
        $nftActivity->from_wallet      = auth()->user()->wallet;
        $nftActivity->to_wallet        = '';
        $nftActivity->type             = 'list';
        $nftActivity->description      = 'Nft listed to marketplace';
        $nftActivity->save();
        return true;
    }

    public function burnNft($id)
    {
        $nft = Nft::where('user_id', auth()->id())->where('status', Status::NFT_ACTIVE)->where('id', $id)->whereHas('category', function ($category) {
            $category->active();
        })->first();

        if (!$nft) {
            return false;
        }

        $nft->status = Status::NFT_BURNED;
        $nft->save();
        return true;
    }

    public function details($slug, $id)
    {
        $pageTitle = 'NFT Details';
        try {
            $id = decrypt($id);
        } catch (\Throwable $th) {
            abort(404);
        }

        $nft = Nft::whereHas('category', function ($category) {
            $category->active();
        })->findOrFail($id);

        if ($nft->status == Status::NFT_BURNED) {
            return to_route('user.nft.all');
        }

        $item = '';

        if ($nft->status == Status::NFT_LISTED) {
            $item = Item::listed()->latest()->with(['bids' => function ($bid) {
                $bid->latest('updated_at');
            }, 'bids.user'])->withCount([
                'wishlists' => function ($wishlist) {
                    $wishlist->where('user_id', auth()->id());
                },
            ])->whereHas('nft.category', function ($category) {
                $category->active();
            })->where('nft_id', $nft->id)->first();

            $this->saveViewCount($nft, $item->id);
        }

        $seoContents['keywords']           = json_decode($nft->meta_keyword);
        $seoContents['social_title']       = $nft->meta_title;
        $seoContents['description']        = $nft->mete_description;
        $seoContents['social_description'] = $nft->meta_description;
        $seoContents['image']              = $nft->file;
        $seoContents                       = (object) $seoContents;
        $seoImage                          = $nft->file;

        return view('Template::user.marketplace.nft_detail', compact('pageTitle', 'nft', 'item', 'item', 'seoContents', 'seoImage'));
    }

    private function saveViewCount($nft, $itemId)
    {
        $nft->increment('views');
        $nftView = NftView::where('nft_id', $nft->id)->whereDate('created_at', today())->first();

        if (!$nftView) {
            $nftView          = new NftView();
            $nftView->nft_id  = $nft->id;
            $nftView->item_id = $itemId;
            $nftView->views   = 1;
            $nftView->save();
        } else {
            $nftView->increment('views');
        }
    }

    public function loadStatistics(Request $request)
    {
        $validator = Validator::make($request->all(), [
            'item_id'  => 'required|integer',
            'duration' => 'required|in:week,month,year',
            'graph'    => 'required|in:views,bids',
        ]);

        if ($validator->fails()) {
            return response()->json(['status' => 'error', 'message' => $validator->errors()]);
        }

        if ($request->graph == 'views') {
            $allData = NftView::where('item_id', $request->item_id);
            $column  = 'views';
        } else {
            $column  = 'amount';
            $allData = Bid::where('item_id', $request->item_id);
        }

        if ($request->duration == 'week') {
            $allData->where('created_at', '>=', now()->startOfWeek())->selectRaw("SUM($column) as cal, DATE_FORMAT(created_at, '%d-%M-%y') as date")->groupBy('date');
        } elseif ($request->duration == 'month') {
            $allData->where('created_at', '>=', now()->startOfMonth())->selectRaw("SUM($column) as cal, DATE_FORMAT(created_at, '%d-%M-%y') as date")->groupBy('date');
        } else {
            $allData->where('created_at', '>=', now()->startOfYear())->selectRaw("SUM($column) as cal, DATE_FORMAT(created_at, '%M-%y') as date")->groupBy('date');
        }

        $allData = $allData->get()->mapWithKeys(function ($data) {
            return [
                $data->date => $data->cal,
            ];
        });

        return response()->json(['status' => 'success', 'data' => $allData]);
    }

    public function nftActivity(Request $request)
    {
        $nftActivities = NftActivity::where('nft_id', $request->nft_id);

        if ($request->activity_type != 'all') {
            $nftActivities->where('type', $request->activity_type);
        }

        if ($request->last_id) {
            $nftActivities->where('id', '<', $request->last_id);
        }

        $nftActivities = $nftActivities->whereHas('nft', function ($nft) {
            $nft->where('status', Status::NFT_LISTED)->orWhere('status', Status::NFT_ACTIVE);
        })->orderBy('id', 'desc')->limit(10)->get();

        return view('Template::partials.nft_activity_list', compact('nftActivities'));
    }

    public function categories()
    {
        $pageTitle  = 'All Categories';
        $categories = Category::active()->paginate(getPaginate());
        $sections   = Page::where('tempname', activeTemplate())->where('slug', 'categories')->first();

        return view('Template::categories', compact('pageTitle', 'categories', 'sections'));
    }

    public function saveMetaInfo(Request $request)
    {
        $request->validate([
            'nft_id'       => 'required|integer',
            'meta_keyword' => 'sometimes|array',
        ]);

        $nft                   = Nft::where('user_id', auth()->id())->findOrFail($request->nft_id);
        $nft->meta_title       = $request->meta_title;
        $nft->meta_description = $request->meta_description;
        $nft->meta_keyword     = $request->meta_keyword;
        $nft->save();

        $notify[] = ['success', 'Nft meta information updated successfully'];
        return back()->withNotify($notify);
    }
}
