<?php

namespace App\Http\Controllers\User;

use App\Http\Controllers\Controller;
use App\Models\Report;
use App\Models\User;
use App\Rules\FileTypeValidate;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Validator;

class ProfileController extends Controller
{
    public function profileSetting()
    {
        $pageTitle = "Profile Setting";
        $user      = auth()->user();

        $info       = json_decode(json_encode(getIpInfo()), true);
        $mobileCode = @implode(',', $info['code']);
        $countries  = json_decode(file_get_contents(resource_path('views/partials/country.json')));

        return view('Template::user.profile_setting', compact('pageTitle', 'user', 'mobileCode', 'countries'));
    }

    public function submitProfile(Request $request)
    {
        $request->validate([
            'profile_image' => ['nullable', 'image', new FileTypeValidate(['jpg', 'jpeg', 'png'])],
            'cover_image'   => ['nullable', 'image', new FileTypeValidate(['jpg', 'jpeg', 'png'])],
            'website_url'   => 'nullable|url',
        ]);

        $existingUser = User::where('id', '!=', auth()->id())->where('email', '!=', null)->where('email', $request->email)->first();

        if ($existingUser) {
            $notify[] = ['error', 'Email already exists'];
            return back()->withNotify($notify);
        }

        $user = auth()->user();

        if ($request->hasFile('profile_image')) {
            try {
                $old                 = $user->profile_image;
                $user->profile_image = fileUploader($request->profile_image, getFilePath('userProfile'), getFileSize('userProfile'), $old);
            } catch (\Exception $exp) {
                $notify[] = ['error', 'Couldn\'t upload your image'];
                return back()->withNotify($notify);
            }
        }

        if ($request->hasFile('cover_image')) {
            try {
                $old               = $user->cover_image;
                $user->cover_image = fileUploader($request->cover_image, getFilePath('userCover'), getFileSize('userCover'), $old, fileManager()->userCover()->thumb);
            } catch (\Exception $exp) {
                $notify[] = ['error', 'Couldn\'t upload your image'];
                return back()->withNotify($notify);
            }
        }

        if (!$user->username) {
            $user->username = $request->username;
        }

        $user->firstname          = $request->firstname;
        $user->lastname           = $request->lastname;
        $user->facebook_username  = $request->facebook_username;
        $user->twitter_username   = $request->twitter_username;
        $user->instagram_username = $request->instagram_username;
        $user->pinterest_username = $request->pinterest_username;
        $user->linkedin_username  = $request->linkedin_username;
        $user->website_url        = $request->website_url;

        $user->save();
        $notify[] = ['success', 'Profile updated successfully'];
        return back()->withNotify($notify);
    }

    public function saveImage(Request $request)
    {
        $validator = Validator::make($request->all(), [
            'image_type'    => 'required|in:profile_image,cover_image',
            'profile_image' => ['required_without:cover_image', 'image', new FileTypeValidate(['jpg', 'jpeg', 'png'])],
            'cover_image'   => ['required_without:profile_image', 'image', new FileTypeValidate(['jpg', 'jpeg', 'png'])],
        ], [
            'profile_image' => 'Profile image is required',
            'cover_image'   => 'Cover image is required',
        ]);

        if ($validator->fails()) {
            return response()->json(['status' => 'error', 'message' => $validator->errors()->all()]);
        }

        $user     = auth()->user();
        $column   = $request->image_type;
        $fileInfo = $column == 'profile_image' ? 'userProfile' : 'userCover';
        $thumb    = $column == 'profile_image' ? null : fileManager()->userCover()->thumb;

        if ($request->hasFile($column)) {
            try {
                $old           = $user->$column;
                $user->$column = fileUploader($request->$column, getFilePath($fileInfo), getFileSize($fileInfo), $old, $thumb);
            } catch (\Exception $exp) {
                return response()->json(['status' => 'error', 'message' => 'Couldn\'t upload your image']);
            }
        }

        $user->save();
        return response()->json(['status' => 'success', 'message' => 'Image updated successfully']);
    }

    public function reportProfile(Request $request)
    {
        $request->validate([
            'user_id' => 'required|integer',
            'quote'   => 'required',
        ]);

        User::findOrFail($request->user_id);

        $report = Report::where('profile_id', $request->user_id)->where('user_id', auth()->id())->first();

        if ($report) {
            $notify[] = ['error', 'You already reported to this profile'];
            return back()->withNotify($notify);
        }

        $report             = new Report();
        $report->profile_id = $request->user_id;
        $report->user_id    = auth()->id();
        $report->quote      = $request->quote;
        $report->save();

        $notify[] = ['success', 'Reported successfully'];
        return back()->withNotify($notify);
    }
}
