<?php

namespace App\Http\Controllers\User;

use App\Constants\Status;
use App\Http\Controllers\Controller;
use App\Lib\FormProcessor;
use App\Lib\GoogleAuthenticator;
use App\Models\DeviceToken;
use App\Models\Follower;
use App\Models\Form;
use App\Models\Frontend;
use App\Models\Transaction;
use App\Models\User;
use App\Models\Wishlist;
use Carbon\Carbon;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Validator;
use Illuminate\Validation\Rule;

class UserController extends Controller
{
    public function kycForm()
    {
        if (auth()->user()->kv == Status::KYC_PENDING) {
            $notify[] = ['error', 'Your KYC is under review'];
            return to_route('user.nft.all')->withNotify($notify);
        }
        if (auth()->user()->kv == Status::KYC_VERIFIED) {
            $notify[] = ['error', 'You are already KYC verified'];
            return to_route('user.nft.all')->withNotify($notify);
        }
        $pageTitle = 'KYC Form';
        $form = Form::where('act', 'kyc')->first();
        return view('Template::user.kyc.form', compact('pageTitle', 'form'));
    }

    public function kycData()
    {
        $user = auth()->user();
        $pageTitle = 'KYC Data';
        return view('Template::user.kyc.info', compact('pageTitle', 'user'));
    }

    public function kycSubmit(Request $request)
    {
        $form = Form::where('act', 'kyc')->firstOrFail();
        $formData = $form->form_data;
        $formProcessor = new FormProcessor();
        $validationRule = $formProcessor->valueValidation($formData);
        $request->validate($validationRule);
        $user = auth()->user();
        foreach (@$user->kyc_data ?? [] as $kycData) {
            if ($kycData->type == 'file') {
                fileManager()->removeFile(getFilePath('verify') . '/' . $kycData->value);
            }
        }
        $userData = $formProcessor->processFormData($request, $formData);
        $user->kyc_data = $userData;
        $user->kyc_rejection_reason = null;
        $user->kv = Status::KYC_PENDING;
        $user->save();

        $notify[] = ['success', 'KYC data submitted successfully'];
        return to_route('user.profile.setting')->withNotify($notify);
    }

    public function addDeviceToken(Request $request)
    {

        $validator = Validator::make($request->all(), [
            'token' => 'required',
        ]);

        if ($validator->fails()) {
            return ['success' => false, 'errors' => $validator->errors()->all()];
        }

        $deviceToken = DeviceToken::where('token', $request->token)->first();

        if ($deviceToken) {
            return ['success' => true, 'message' => 'Already exists'];
        }

        $deviceToken          = new DeviceToken();
        $deviceToken->user_id = auth()->user()->id;
        $deviceToken->token   = $request->token;
        $deviceToken->is_app  = Status::NO;
        $deviceToken->save();

        return ['success' => true, 'message' => 'Token saved successfully'];
    }

    public function downloadAttachment($fileHash)
    {
        $filePath = decrypt($fileHash);
        $extension = pathinfo($filePath, PATHINFO_EXTENSION);
        $title = slug(gs('site_name')) . '- attachments.' . $extension;
        try {
            $mimetype = mime_content_type($filePath);
        } catch (\Exception $e) {
            $notify[] = ['error', 'File does not exists'];
            return back()->withNotify($notify);
        }
        header('Content-Disposition: attachment; filename="' . $title);
        header("Content-Type: " . $mimetype);
        return readfile($filePath);
    }

    public function authorizeForm()
    {
        $user = auth()->user();
        if ($user->status) {
            return to_route('user.nft.all');
        }
        $pageTitle = 'Banned';
        $banContent = Frontend::where('data_keys', 'ban_page.content')->first();
        return view('Template::user.authorization.ban', compact('user', 'banContent', 'pageTitle'));
    }

    public function wishlist()
    {
        $pageTitle = 'My Wishlists';
        $wishlists = Wishlist::with('item.bids.user', 'item.user', 'item.nft')->where('user_id', auth()->id())->paginate(getPaginate());
        $user      = auth()->user();

        return view('Template::user.wishlist', compact('pageTitle', 'wishlists', 'user'));
    }

    public function follower()
    {
        $pageTitle = 'Follower List';
        $user      = auth()->user();

        $followers      = Follower::with('followerUser')->where('user_id', $user->id)->get();
        $myFollowingIds = Follower::where('follower_id', $user->id)->pluck('user_id')->toArray();

        return view('Template::user.follower', compact('pageTitle', 'user', 'followers', 'myFollowingIds'));
    }
    public function following()
    {
        $pageTitle = 'Following List';
        $user      = auth()->user();

        $followings = Follower::with('followingUser')->where('follower_id', $user->id)->get();

        return view('Template::user.following', compact('pageTitle', 'user', 'followings'));
    }
    public function sendEmailOtp(Request $request)
    {
        $validator = Validator::make($request->all(), [
            'email' => 'required|email',
        ]);

        if ($validator->fails()) {
            return response()->json(['status' => 'error', 'message' => $validator->errors()->all()]);
        }

        $user = auth()->user();

        if ($user->ev == 1) {
            return response()->json(['status' => 'error', 'message' => 'You already verified your email']);
        }

        $exists = User::where('id', '!=', $user->id)->where('email', $request->email)->first();

        if ($exists) {
            return response()->json(['status' => 'error', 'message' => 'Email already exists']);
        }

        if ($user->email_ver_code_send_at && Carbon::parse($user->email_ver_code_send_at)->addMinutes(2) > now()) {
            $nextTime = Carbon::parse($user->email_ver_code_send_at)->addMinutes(2)->diffInSeconds(now());
            return response()->json(['status' => 'error', 'message' => "You can request a code again after $nextTime seconds"]);
        }

        $user->email_ver_code         = verificationCode(6);
        $user->email_ver_code_send_at = now();
        $user->save();

        session()->put('user_email', $request->email);

        notify($user, 'EVER_CODE', [
            'code' => $user->email_ver_code,
        ], ['email']);

        return response()->json(['status' => 'success', 'message' => 'Verification code send successfully']);
    }
    public function verifyOtp(Request $request)
    {
        $validator = Validator::make($request->all(), [
            'otp_code' => 'required',
            'otp_type' => 'required|in:mobile,email',
        ]);

        if ($validator->fails()) {
            return response()->json(['status' => 'error', 'message' => $validator->errors()->all()]);
        }

        if ($request->otp_type == 'email') {
            $userEmail = session()->get('user_email');
            if (!$userEmail) {
                return response()->json(['status' => 'error', 'message' => 'Session Timeout']);
            }
        } else {
            $userMobile = session()->get('user_mobile');
            if (!$userMobile) {
                return response()->json(['status' => 'error', 'message' => 'Session Timeout']);
            }
        }

        $user = auth()->user();

        if ($request->otp_type == 'mobile' && @$user->mobile_ver_code == $request->otp_code) {
            $user->mobile                  = $userMobile;
            $user->sv                      = Status::VERIFIED;
            $user->mobile_ver_code         = null;
            $user->mobile_ver_code_send_at = null;
        } elseif (@$user->email_ver_code == $request->otp_code) {
            $user->email                  = $userEmail;
            $user->ev                     = Status::VERIFIED;
            $user->email_ver_code         = null;
            $user->email_ver_code_send_at = null;
        } else {
            return response()->json(['status' => 'error', 'message' => 'Verification code didn\'t match!']);
        }

        $user->save();
        return response()->json(['status' => 'success', 'message' => ucfirst($request->otp_type) . ' verified successfully']);
    }
    public function sendMobileOtp(Request $request)
    {
        $countryData  = (array) json_decode(file_get_contents(resource_path('views/partials/country.json')));
        $countryCodes = implode(',', array_keys($countryData));
        $mobileCodes  = implode(',', array_column($countryData, 'dial_code'));

        $validator = Validator::make($request->all(), [
            'mobile'       => 'required',
            'mobile_code'  => 'required|in:' . $mobileCodes,
            'country_code' => 'required|in:' . $countryCodes,
        ]);

        if ($validator->fails()) {
            return response()->json(['status' => 'error', 'message' => $validator->errors()->all()]);
        }

        $user = auth()->user();

        if ($user->sv == 1) {
            return response()->json(['status' => 'error', 'message' => 'You already verified your mobile']);
        }

        $mobile = $request->mobile_code . $request->mobile;

        $exists = User::where('id', '!=', $user->id)->where('mobile', $mobile)->first();

        if ($exists) {
            return response()->json(['status' => 'error', 'message' => 'Mobile already exists']);
        }

        if ($user->mobile_ver_code_send_at && Carbon::parse($user->mobile_ver_code_send_at)->addMinutes(2) > now()) {
            $nextTime = Carbon::parse($user->mobile_ver_code_send_at)->addMinutes(2)->diffInSeconds(now());
            return response()->json(['status' => 'error', 'message' => "You can request a code again after $nextTime seconds"]);
        }

        $user->country_code            = $request->country_code;
        $user->mobile_ver_code         = verificationCode(6);
        $user->mobile_ver_code_send_at = now();
        $user->save();

        session()->put('user_mobile', $mobile);

        notify($user, 'SVER_CODE', [
            'code' => $user->mobile_ver_code,
        ], ['sms']);

        return response()->json(['status' => 'success', 'message' => 'Verification code send successfully']);
    }
}
