// SPDX-License-Identifier: MIT
// File: @openzeppelin/contracts/utils/introspection/IERC165.sol


// OpenZeppelin Contracts (last updated v5.0.0) (utils/introspection/IERC165.sol)

pragma solidity ^0.8.20;

/**
 * @dev Interface of the ERC165 standard, as defined in the
 * https://eips.ethereum.org/EIPS/eip-165[EIP].
 *
 * Implementers can declare support of contract interfaces, which can then be
 * queried by others ({ERC165Checker}).
 *
 * For an implementation, see {ERC165}.
 */
interface IERC165 {
    /**
     * @dev Returns true if this contract implements the interface defined by
     * `interfaceId`. See the corresponding
     * https://eips.ethereum.org/EIPS/eip-165#how-interfaces-are-identified[EIP section]
     * to learn more about how these ids are created.
     *
     * This function call must use less than 30 000 gas.
     */
    function supportsInterface(bytes4 interfaceId) external view returns (bool);
}

// File: @openzeppelin/contracts/token/ERC721/IERC721.sol


// OpenZeppelin Contracts (last updated v5.0.0) (token/ERC721/IERC721.sol)

pragma solidity ^0.8.20;


/**
 * @dev Required interface of an ERC721 compliant contract.
 */
interface IERC721 is IERC165 {
    /**
     * @dev Emitted when `tokenId` token is transferred from `from` to `to`.
     */
    event Transfer(address indexed from, address indexed to, uint256 indexed tokenId);

    /**
     * @dev Emitted when `owner` enables `approved` to manage the `tokenId` token.
     */
    event Approval(address indexed owner, address indexed approved, uint256 indexed tokenId);

    /**
     * @dev Emitted when `owner` enables or disables (`approved`) `operator` to manage all of its assets.
     */
    event ApprovalForAll(address indexed owner, address indexed operator, bool approved);

    /**
     * @dev Returns the number of tokens in ``owner``'s account.
     */
    function balanceOf(address owner) external view returns (uint256 balance);

    /**
     * @dev Returns the owner of the `tokenId` token.
     *
     * Requirements:
     *
     * - `tokenId` must exist.
     */
    function ownerOf(uint256 tokenId) external view returns (address owner);

    /**
     * @dev Safely transfers `tokenId` token from `from` to `to`.
     *
     * Requirements:
     *
     * - `from` cannot be the zero address.
     * - `to` cannot be the zero address.
     * - `tokenId` token must exist and be owned by `from`.
     * - If the caller is not `from`, it must be approved to move this token by either {approve} or {setApprovalForAll}.
     * - If `to` refers to a smart contract, it must implement {IERC721Receiver-onERC721Received}, which is called upon
     *   a safe transfer.
     *
     * Emits a {Transfer} event.
     */
    function safeTransferFrom(address from, address to, uint256 tokenId, bytes calldata data) external;

    /**
     * @dev Safely transfers `tokenId` token from `from` to `to`, checking first that contract recipients
     * are aware of the ERC721 protocol to prevent tokens from being forever locked.
     *
     * Requirements:
     *
     * - `from` cannot be the zero address.
     * - `to` cannot be the zero address.
     * - `tokenId` token must exist and be owned by `from`.
     * - If the caller is not `from`, it must have been allowed to move this token by either {approve} or
     *   {setApprovalForAll}.
     * - If `to` refers to a smart contract, it must implement {IERC721Receiver-onERC721Received}, which is called upon
     *   a safe transfer.
     *
     * Emits a {Transfer} event.
     */
    function safeTransferFrom(address from, address to, uint256 tokenId) external;

    /**
     * @dev Transfers `tokenId` token from `from` to `to`.
     *
     * WARNING: Note that the caller is responsible to confirm that the recipient is capable of receiving ERC721
     * or else they may be permanently lost. Usage of {safeTransferFrom} prevents loss, though the caller must
     * understand this adds an external call which potentially creates a reentrancy vulnerability.
     *
     * Requirements:
     *
     * - `from` cannot be the zero address.
     * - `to` cannot be the zero address.
     * - `tokenId` token must be owned by `from`.
     * - If the caller is not `from`, it must be approved to move this token by either {approve} or {setApprovalForAll}.
     *
     * Emits a {Transfer} event.
     */
    function transferFrom(address from, address to, uint256 tokenId) external;

    /**
     * @dev Gives permission to `to` to transfer `tokenId` token to another account.
     * The approval is cleared when the token is transferred.
     *
     * Only a single account can be approved at a time, so approving the zero address clears previous approvals.
     *
     * Requirements:
     *
     * - The caller must own the token or be an approved operator.
     * - `tokenId` must exist.
     *
     * Emits an {Approval} event.
     */
    function approve(address to, uint256 tokenId) external;

    /**
     * @dev Approve or remove `operator` as an operator for the caller.
     * Operators can call {transferFrom} or {safeTransferFrom} for any token owned by the caller.
     *
     * Requirements:
     *
     * - The `operator` cannot be the address zero.
     *
     * Emits an {ApprovalForAll} event.
     */
    function setApprovalForAll(address operator, bool approved) external;

    /**
     * @dev Returns the account approved for `tokenId` token.
     *
     * Requirements:
     *
     * - `tokenId` must exist.
     */
    function getApproved(uint256 tokenId) external view returns (address operator);

    /**
     * @dev Returns if the `operator` is allowed to manage all of the assets of `owner`.
     *
     * See {setApprovalForAll}
     */
    function isApprovedForAll(address owner, address operator) external view returns (bool);
}

// File: @openzeppelin/contracts/security/ReentrancyGuard.sol


// OpenZeppelin Contracts (last updated v4.9.0) (security/ReentrancyGuard.sol)

pragma solidity ^0.8.0;

/**
 * @dev Contract module that helps prevent reentrant calls to a function.
 *
 * Inheriting from `ReentrancyGuard` will make the {nonReentrant} modifier
 * available, which can be applied to functions to make sure there are no nested
 * (reentrant) calls to them.
 *
 * Note that because there is a single `nonReentrant` guard, functions marked as
 * `nonReentrant` may not call one another. This can be worked around by making
 * those functions `private`, and then adding `external` `nonReentrant` entry
 * points to them.
 *
 * TIP: If you would like to learn more about reentrancy and alternative ways
 * to protect against it, check out our blog post
 * https://blog.openzeppelin.com/reentrancy-after-istanbul/[Reentrancy After Istanbul].
 */
abstract contract ReentrancyGuard {
    // Booleans are more expensive than uint256 or any type that takes up a full
    // word because each write operation emits an extra SLOAD to first read the
    // slot's contents, replace the bits taken up by the boolean, and then write
    // back. This is the compiler's defense against contract upgrades and
    // pointer aliasing, and it cannot be disabled.

    // The values being non-zero value makes deployment a bit more expensive,
    // but in exchange the refund on every call to nonReentrant will be lower in
    // amount. Since refunds are capped to a percentage of the total
    // transaction's gas, it is best to keep them low in cases like this one, to
    // increase the likelihood of the full refund coming into effect.
    uint256 private constant _NOT_ENTERED = 1;
    uint256 private constant _ENTERED = 2;

    uint256 private _status;

    constructor() {
        _status = _NOT_ENTERED;
    }

    /**
     * @dev Prevents a contract from calling itself, directly or indirectly.
     * Calling a `nonReentrant` function from another `nonReentrant`
     * function is not supported. It is possible to prevent this from happening
     * by making the `nonReentrant` function external, and making it call a
     * `private` function that does the actual work.
     */
    modifier nonReentrant() {
        _nonReentrantBefore();
        _;
        _nonReentrantAfter();
    }

    function _nonReentrantBefore() private {
        // On the first call to nonReentrant, _status will be _NOT_ENTERED
        require(_status != _ENTERED, "ReentrancyGuard: reentrant call");

        // Any calls to nonReentrant after this point will fail
        _status = _ENTERED;
    }

    function _nonReentrantAfter() private {
        // By storing the original value once again, a refund is triggered (see
        // https://eips.ethereum.org/EIPS/eip-2200)
        _status = _NOT_ENTERED;
    }

    /**
     * @dev Returns true if the reentrancy guard is currently set to "entered", which indicates there is a
     * `nonReentrant` function in the call stack.
     */
    function _reentrancyGuardEntered() internal view returns (bool) {
        return _status == _ENTERED;
    }
}

// File: contracts/Event.sol


pragma solidity ^0.8.26;

contract Event  {

    event Offered(
        uint256 itemId,
        address indexed nft,
        uint256 tokenId,
        uint256 price,
        bool isAuction,
        address indexed seller
    );
    event Unlisted(
        uint256 itemId,
        address indexed nft,
        uint256 tokenId,
        address indexed seller
    );

    event Bought(
        uint256 itemId,
        address indexed nft,
        uint256 tokenId,
        uint256 price,
        address indexed seller,
        address indexed buyer
    );

    event BidHistory(
        uint256 itemId,
        address indexed nft,
        uint256 tokenId,
        address indexed bidder,
        uint256 amount,
        string remark,
        uint256 timestamp
    );

    event BidWinner(
        uint256 itemId,
        address indexed nft,
        uint256 tokenId,
        uint256 amount,
        address indexed bidder,
        address indexed seller,
        uint256 timestamp
    );

    event AuctionBack(
        uint256 itemId,
        address indexed nft,
        uint256 tokenId,
        address indexed seller,
        uint256 timestamp
    );
}

// File: contracts/Users.sol


pragma solidity ^0.8.26;

contract Users {
    struct User {
        address wallet;
        uint256 balance;
        uint256 total_deposits;
        uint256 total_sell;
        uint256 total_withdraw;
    }

    event Transaction(
        address wallet,
        uint256 amount,
        bytes32 trx_type,
        uint256 post_balance,
        uint256 charge,
        uint256 timestamp
    );

    mapping(address => User) users;

    function addUser() external {
        if (users[msg.sender].wallet == address(0)) {
            users[msg.sender] = User(msg.sender, 0, 0, 0, 0);
        }
    }
    function findOraddUser() internal returns(User storage) {
        User storage _user = users[msg.sender];
        if (_user.wallet == address(0)) {
            users[msg.sender] = User(msg.sender, 0, 0, 0, 0);
        }
        return(users[msg.sender]);
    }

    function getUser() external view returns (User memory) {
        return (users[msg.sender]);
    }

    function updateTotalDeposit(uint256 _charge) internal {
        User storage _user = users[msg.sender];
        updateBalance(_user, (msg.value - _charge), _charge, "+");
        _user.total_deposits += (msg.value - _charge);
    }

    function updateTotalWithdrawal(uint256 _amount, uint256 _charge) internal {
        User storage _user = users[msg.sender];
        updateBalance(_user, (_amount), _charge, "-");
        _user.total_withdraw += (_amount);
    }

    function updateBalance(
        User storage _user,
        uint256 _amount,
        uint256 _charge,
        bytes32 _trxType
    ) internal {
        if (_trxType == "+") {
            _user.balance += _amount;
        }else{
            _user.balance -= _amount;
        }
        emit Transaction(
            msg.sender,
            _amount,
            _trxType,
            _user.balance,
            _charge,
            block.timestamp
        );
    }
}

// File: contracts/Deposits.sol


pragma solidity ^0.8.26;

contract Deposits {
    uint256 public immutable percentDepositCharge;

    struct Deposit {
        address wallet;
        uint256 amount;
        uint256 charge;
        uint256 withCharge;
        uint256 timestamp;
    }
    mapping(address => Deposit[]) deposits;
    modifier valueNonZero() {
        require(msg.value > 0, "Please send greater than zero.");
        _;
    }

    constructor(uint256 _percentDepositCharge) {
        percentDepositCharge = _percentDepositCharge;
    }

    function addDeposit(uint256 _charge) internal valueNonZero {
        Deposit[] storage _deposits = deposits[msg.sender];
        Deposit memory _deposit = Deposit(
            msg.sender,
            msg.value - _charge,
            _charge,
            msg.value,
            block.timestamp
        );
        _deposits.push(_deposit);
    }

    function getDepositByCaller() external view returns (Deposit[] memory) {
        return (deposits[msg.sender]);
    }

    function getDepostisCountByCaller() external view returns (uint256) {
        return (deposits[msg.sender].length);
    }
}

// File: contracts/Withdrawals.sol


pragma solidity ^0.8.26;

contract Withdarawals {
    uint256 public immutable percentWithdrawalCharge;

    struct Withdarawal {
        address wallet;
        uint256 amount;
        uint256 charge;
        uint256 withCharge;
        uint256 timestamp;
    }

    mapping(address => Withdarawal[]) withdarawals;

    constructor(uint256 _percentWithdrawalCharge) {
        percentWithdrawalCharge = _percentWithdrawalCharge;
    }

    function addWithdrawal(
        uint256 _amount,
        uint256 _charge,
        address userWallet
    ) internal {
        payable(userWallet).transfer(_amount - _charge);
        Withdarawal[] storage _withdrawals = withdarawals[msg.sender];
        Withdarawal memory _withdrawal = Withdarawal(
            msg.sender,
            _amount - _charge,
            _charge,
            _amount,
            block.timestamp
        );
        _withdrawals.push(_withdrawal);
    }

    function getWithdrawalByCaller()
        external
        view
        returns (Withdarawal[] memory)
    {
        return (withdarawals[msg.sender]);
    }

    function getWithdrawalsCountByCaller() external view returns (uint256) {
        return (withdarawals[msg.sender].length);
    }
}

// File: contracts/Marketplace.sol


pragma solidity ^0.8.26;







contract Marketplace is ReentrancyGuard, Event, Users, Deposits, Withdarawals {
    address payable immutable feeAccount;
    uint256 public immutable feePercent;

    uint256 public itemCount;

    mapping(uint256 => uint256) public heighestBid; // (itemId => amount)
    mapping(uint256 => address) public heighestBidder; // (itemId => bidder address)

    enum ItemStatus {
        Unlisted,
        Sold,
        Listed
    }

    struct Item {
        uint256 itemId;
        IERC721 nft;
        uint256 tokenId;
        uint256 price;
        address payable seller;
        ItemStatus status;
        bool isAuction;
        uint256 endTime;
    }

    mapping(uint256 => Item) public items; // (itemId => Item)

    struct Bid {
        uint256 itemId;
        address bidder;
        uint256 amount;
        bool isWinner;
        uint256 timestamp;
    }
    mapping(uint256 => mapping(address => Bid)) public bids; // (itemId => (bidder address => Bid))

    constructor(
        uint256 _feePercent,
        uint256 _percentDepositCharge,
        uint256 _percentWithdrawalCharge
    ) Deposits(_percentDepositCharge) Withdarawals(_percentWithdrawalCharge) {
        feeAccount = payable(msg.sender);
        feePercent = _feePercent;
    }

    function makeItem(
        IERC721 _nft,
        uint256 _tokenId,
        uint256 _price,
        uint256 _endTime
    ) external nonReentrant {
        findOraddUser();
        require(_price > 0, "Price must be greater than zero");
        bool _isAuction = _endTime > 0 ? true : false;

        _nft.transferFrom(msg.sender, address(this), _tokenId);

        itemCount++;

        items[itemCount] = Item(
            itemCount,
            _nft,
            _tokenId,
            _price,
            payable(msg.sender),
            ItemStatus.Listed,
            _isAuction,
            _endTime
        );

        emit Offered(
            itemCount,
            address(_nft),
            _tokenId,
            _price,
            _isAuction,
            msg.sender
        );
    }

    function unlistItem(uint256 _itemId) public {
        Item storage item = items[_itemId];
        require(item.seller == msg.sender, "Caller must be owner of this item");
        require(item.status != ItemStatus.Sold, "Item already sold");
        require(item.status != ItemStatus.Unlisted, "Item Unlisted by seller");
        require(
            item.isAuction == false || item.endTime < block.timestamp,
            "Unlist failed! Auction is still running"
        );

        item.nft.transferFrom(address(this), item.seller, item.tokenId);
        emit Unlisted(_itemId, address(item.nft), item.tokenId, item.seller);
    }

    function purchaseItem(uint256 _itemId, bool _fromBalance)
        external
        payable
        nonReentrant
    {
        Item storage item = items[_itemId];
        require(
            item.itemId > 0 && item.itemId <= itemCount,
            "Item doesn't exists"
        );
        User storage _user = findOraddUser();
        require(
            !item.isAuction,
            "The auction item cannot be purchased directly."
        );

        require(item.status != ItemStatus.Sold, "Item already sold");
        require(item.status != ItemStatus.Unlisted, "Item Unlisted by seller");
        uint256 _itemPrice = item.price;

        if (_fromBalance) {
            require(_user.balance >= _itemPrice, "Insufficient Balance");
            updateBalance(_user, _itemPrice, 0, "-");
        } else {
            require(
                msg.value == _itemPrice,
                "Please send the exact amount to purchase the NFT"
            );
        }

        uint256 _tradingCharge = getCharge(_itemPrice);
        item.seller.transfer(_itemPrice - _tradingCharge);
        feeAccount.transfer(_tradingCharge);
        item.status = ItemStatus.Sold;
        item.nft.transferFrom(address(this), msg.sender, item.tokenId);

        emit Bought(
            item.itemId,
            address(item.nft),
            item.tokenId,
            item.price,
            item.seller,
            msg.sender
        );
    }

    function getItemPrice(uint256 _itemId) public view returns (uint256) {
        Item memory item = items[_itemId];
        return item.price;
    }

    function bidItem(
        uint256 _itemId,
        bool _fromBalance,
        uint256 _bidAmountFromBalance
    ) external payable nonReentrant {
        Item storage item = items[_itemId];
        require(
            item.itemId > 0 && item.itemId <= itemCount,
            "item doesn't exists"
        );
        User storage _user = findOraddUser();

        require(item.isAuction, "It is not an auction item");
        require(item.status != ItemStatus.Sold, "Item already sold");
        require(item.status != ItemStatus.Unlisted, "Item Unlisted by seller");
        require(
            item.endTime > block.timestamp,
            "Auction is not running currently"
        );

        Bid storage _oldBid = bids[item.itemId][msg.sender];
        require(_oldBid.amount == 0, "Update bid insted of placing a new bid.");

        uint256 _bidAmount = _fromBalance ? _bidAmountFromBalance : msg.value;
        require(_bidAmount >= item.price, "Please follow minimum bid amount");

        uint256 _heighestBid = heighestBid[item.itemId];
        require(
            _bidAmount > _heighestBid,
            "Bid must be greater than latest heighest bid"
        );

        if (_fromBalance) {
            require(
                _user.balance >= _bidAmountFromBalance,
                "Insufficiant Balance"
            );
            updateBalance(_user, _bidAmount, 0, "-");
        }

        heighestBid[item.itemId] = _bidAmount;
        heighestBidder[item.itemId] = msg.sender;

        Bid memory _bid = Bid(
            item.itemId,
            msg.sender,
            _bidAmount,
            false,
            block.timestamp
        );

        bids[item.itemId][msg.sender] = _bid;

        emit BidHistory(
            item.itemId,
            address(item.nft),
            item.tokenId,
            msg.sender,
            _bidAmount,
            "Bid placed",
            block.timestamp
        );
    }

    function updateBid(
        uint256 _itemId,
        bool _fromBalance,
        uint256 _bidAmountFromBalance
    ) external payable nonReentrant {
        Item storage item = items[_itemId];
        require(
            item.itemId > 0 && item.itemId <= itemCount,
            "item doesn't exists"
        );
        User storage _user = findOraddUser();
        require(item.isAuction, "It is not an auction item");
        require(item.status != ItemStatus.Sold, "Item already sold");
        require(item.status != ItemStatus.Unlisted, "Item Unlisted by seller");
        require(
            item.endTime > block.timestamp,
            "Auction is not running currently"
        );

        Bid storage _bid = bids[item.itemId][msg.sender];
        require(_bid.amount > 0, "No bid found for this item");

        uint256 _heighestBid = heighestBid[item.itemId];
        uint256 _additionalBidAmount = _fromBalance
            ? _bidAmountFromBalance
            : msg.value;
        require(
            (_bid.amount + _additionalBidAmount) > _heighestBid,
            "Bid amount must be greater than latest heighest bid"
        );
        if (_fromBalance) {
            require(
                _bidAmountFromBalance <= _user.balance,
                "Insufficiant Balance"
            );
            updateBalance(_user, _additionalBidAmount, 0, "-");
        }
        _bid.amount += _additionalBidAmount;
        heighestBid[item.itemId] = _bid.amount;
        heighestBidder[item.itemId] = msg.sender;

        emit BidHistory(
            item.itemId,
            address(item.nft),
            item.tokenId,
            msg.sender,
            _bid.amount,
            "Bid updated",
            block.timestamp
        );
    }

    function claimBidAmount(uint256 _itemId, bool _toBalance)
        external
        payable
        nonReentrant
    {
        Item storage item = items[_itemId];
        require(
            item.itemId > 0 && item.itemId <= itemCount,
            "item doesn't exists"
        );
        User storage _user = findOraddUser();
        address _heighestBidder = heighestBidder[item.itemId];
        require(
            _heighestBidder != msg.sender,
            "You cant claim bid amount, as you are the highest bidder"
        );
        Bid storage _bid = bids[item.itemId][msg.sender];
        uint256 _bidAmount = _bid.amount;
        require(_bidAmount > 0, "You have no bid yet");
        if (_toBalance) {
            updateBalance(_user, _bidAmount, 0, "+");
        } else {
            payable(msg.sender).transfer(_bidAmount);
        }
        _bid.amount = 0;
    }

    function getBidByItemId(uint256 _itemId)
        external
        view
        returns (Bid memory)
    {
        return (bids[_itemId][msg.sender]);
    }

    function makeWinner(uint256 _itemId) external payable {
        Item storage item = items[_itemId];
        require(item.status != ItemStatus.Sold, "Item already sold");
        require(item.status != ItemStatus.Unlisted, "Item Unlisted by seller");
        require(item.endTime < block.timestamp, "Auction is still running");
        uint256 _heighestBid = heighestBid[item.itemId];
        if (_heighestBid > 0) {
            address _heighestBidder = heighestBidder[item.itemId];
            uint256 _tradingCharge = getCharge(_heighestBid);

            item.seller.transfer(_heighestBid - _tradingCharge);

            feeAccount.transfer(_tradingCharge);

            item.status = ItemStatus.Sold;
            item.nft.transferFrom(address(this), _heighestBidder, item.tokenId);
            Bid storage _bid = bids[item.itemId][msg.sender];
            _bid.amount = 0;
            emit BidWinner(
                _itemId,
                address(item.nft),
                item.tokenId,
                _heighestBid,
                _heighestBidder,
                item.seller,
                block.timestamp
            );
        } else {
            item.nft.transferFrom(address(this), item.seller, item.tokenId);
            emit AuctionBack(
                _itemId,
                address(item.nft),
                item.tokenId,
                item.seller,
                block.timestamp
            );
        }
    }

    function withCharge(uint256 _amount) private view returns (uint256) {
        return (_amount + ((_amount * feePercent) / 100));
    }

    function getCharge(uint256 _amount) private view returns (uint256) {
        return ((_amount * feePercent) / 100);
    }

    function getBidAmount(uint256 _amount) public view returns (uint256) {
        return ((100 * _amount) / (100 + feePercent));
    }

    function getTotalPrice(uint256 _itemId) public view returns (uint256) {
        return ((items[_itemId].price * (100 + feePercent)) / 100);
    }

    function depositNow() external payable {
        findOraddUser();
        uint256 _charge = ((msg.value * percentDepositCharge) / 100);
        addDeposit(_charge);
        feeAccount.transfer(_charge);
        updateTotalDeposit(_charge);
    }

    function withdrawNow(uint256 _amount) external payable {
        _amount = _amount;
        User storage user = findOraddUser();
        require(user.balance > _amount, "Insufficient balance");
        uint256 _charge = ((_amount * percentWithdrawalCharge) / 100);
        addWithdrawal(_amount, _charge, msg.sender);
        feeAccount.transfer(_charge);
        updateTotalWithdrawal(_amount, _charge);
    }

    function withdrawContractBalance() external payable {
        require(
            feeAccount == msg.sender,
            "Caller must be owner of this contract"
        );
        feeAccount.transfer(address(this).balance);
    }

    function getBalance() public view returns (uint256) {
        require(
            feeAccount == msg.sender,
            "Caller must be owner of this contract"
        );
        return address(this).balance;
    }
}
